"""The module provides interpolators."""

from __future__ import annotations

import logging
from typing import Any

import numpy as np
import numpy.typing as npt
from scipy.interpolate import LinearNDInterpolator, NearestNDInterpolator

logger = logging.getLogger(__name__)


class Interpolator2D:
    """Two-dimensional interpolator."""

    def __init__(
        self,
        x: npt.NDArray[np.float64],
        y: npt.NDArray[np.float64],
        z: npt.NDArray[np.float64],
        kind: str = "linear",
        fill_value: float = np.nan,
    ):
        """

        Parameters
        ----------
        x :
            X-axis values (1-D array)
        y :
            Y-axis values (1-D array)
        z :
            Data values (1-D array)
        kind : {'nearest', 'linear'}
            The kind of interpolation to use.
        fill_value :
            Value used to fill in for requested points outside of the convex hull
            of the input points.

        """
        self.x = x
        self.y = y
        self.z = z

        self.kind = kind
        self.fill_value = fill_value

        self._interpolator = None

    @property
    def interpolator(self):
        """The actual interpolator."""
        if self._interpolator is None:
            if self.kind == "nearest":
                interp = NearestNDInterpolator((self.x, self.y), self.z)
            elif self.kind == "linear":
                interp = LinearNDInterpolator((self.x, self.y), self.z, self.fill_value)
            else:
                raise ValueError(f"Unknown interpolation method: {self.kind}")
            self._interpolator = interp
        return self._interpolator

    def update(self, parameters: dict[str, Any]):
        """Update the interpolator parameters."""
        for key, value in parameters.items():
            setattr(self, key, value)
        logger.info("The interpolator parameters have been updated.")
        self._interpolator = None

    def __call__(self, points: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """Evaluate interpolator."""
        return self.interpolator.__call__(points)



"""The module provides classes to deal with different types of measurements."""

from __future__ import annotations

import contextlib
import copy
import logging
from abc import ABC, abstractmethod
from functools import cached_property
from itertools import cycle
from typing import Any, Iterable

import numpy as np
import numpy.typing as npt

from daxs.interpolators import Interpolator2D
from daxs.scans import Scan, Scans
from daxs.sources import Source
from daxs.utils.arrays import trapezoid

logger = logging.getLogger(__name__)


class ConcentrationCorrectionError(Exception):
    def __init__(self, message: str):
        super().__init__(f"{message} The concentration correction failed.")


class Corrector(ABC):
    """Base class for measurement correctors."""

    @abstractmethod
    def apply(self, scans: Scans) -> None:
        """Apply the correction to the scans."""


class SimpleConcentrationCorrector(Corrector):
    """Class to perform simple, length-based, concentration corrections."""

    def __init__(self, scans: Scan | Scans):
        """
        Parameters
        ----------
        scans :
            Scans used for concentration correction.
        """
        self.conc_corr_scans = scans if isinstance(scans, Scans) else Scans([scans])

    def apply(self, scans: Scans) -> None:
        logger.info("Applying simple concentration correction.")
        # When there is a single concentration correction scan and the number
        # of points in it is equal to the number of scans to be corrected, each
        # point will be used to correct a scan.
        if len(self.conc_corr_scans) == 1:
            [conc_corr_scan] = self.conc_corr_scans
            if len(scans) == conc_corr_scan.signal.size:
                for i, scan in enumerate(scans):
                    try:
                        scalars = (conc_corr_scan.signal[i], conc_corr_scan.monitor[i])
                    except IndexError:
                        scalars = (conc_corr_scan.signal[i],)
                    scan.divide_by_scalars(*scalars)
                return

        # When there is a single concentration correction scan and the previous
        # condition is not met, divide all scans by it, by cycling it.
        if len(self.conc_corr_scans) == 1:
            conc_corr_scans = cycle(self.conc_corr_scans)

        # When the number of scans to be corrected is equal to the number of
        # concentration correction scans, each scan will be corrected by a
        # corresponding concentration correction scan.
        elif len(self.conc_corr_scans) == len(scans):
            conc_corr_scans = self.conc_corr_scans
        # No other case is supported.
        else:
            raise ConcentrationCorrectionError(
                "Incompatible number of scans to correct and concentration "
                "correction scans."
            )

        for scan, conc_corr_scan in zip(scans, conc_corr_scans):
            try:
                scan.divide_by_scan(conc_corr_scan)
            except (TypeError, ValueError) as e:
                raise ConcentrationCorrectionError(
                    f"The length of the signal or monitor in the scan {scan.label} "
                    "is different than that from the correction scan "
                    f"{conc_corr_scan.label}."
                ) from e


class DataDrivenConcentrationCorrector(Corrector):
    """Class to perform concentration corrections using data from specified paths."""

    def __init__(self, scans: Scan | Scans, paths: dict[str, str]):
        """
        Parameters
        ----------
        scans :
            Scans used for concentration correction.
        paths :
            Paths to the concentration correction data. It assumes they are the
            same for specified scans.
        """
        self.conc_corr_scans = scans if isinstance(scans, Scans) else Scans([scans])
        self.paths = paths

    @cached_property
    def conc_corr_points(self) -> npt.NDArray[np.float64]:
        """
        Array of points used to determine the concentration correction indices.

        Returns
        -------
        npt.NDArray[np.float64]
            Array of points used for concentration correction.

        Raises
        ------
        ValueError
            If the concentration correction counters don't have the same length.
        """
        points = []
        for path in self.paths.values():
            points_at_path = []
            for conc_corr_scan in self.conc_corr_scans:
                points_at_path.extend(conc_corr_scan.read_data_at_paths(path))
            points.append(points_at_path)

        try:
            return np.asarray(points, dtype=np.float64).T
        except ValueError as e:
            raise ConcentrationCorrectionError(
                "The concentration correction counters must have the same length."
            ) from e

    def find_conc_corr_indices(self, scan: Scan) -> list[int]:
        """
        Determine the indices of the concentration correction data for each point in
        the scan.

        Parameters
        ----------
        scan :
            Scan for which the concentration correction data indices are to be found.

        Returns
        -------
        list[int]
            Indices of the concentration correction data for the points in the scan.
        """
        # Get concentration correction points.
        conc_corr_points = self.conc_corr_points
        # conc_corr_points[-1, :] = conc_corr_points[80, :]

        # Get the scan data at the same keys as the concentration correction data.
        data_points = []
        for key in self.paths:
            try:
                data_points.append(scan.data[key])
            except KeyError as e:
                raise ConcentrationCorrectionError(
                    f"The data in scan {scan.label} does not have the key {key} among"
                    "the source data paths. Make sure the source data paths are"
                    "correctly set."
                ) from e
        # data_points = [
        #     [8.6775, 8.82, 8.9625],
        #     [-10.63, -10.63, -10.63],
        # ]
        data_points = np.asarray(data_points)
        # Add a new axis if the data points are 1D.
        if data_points.ndim == 1:
            data_points = data_points[:, None]
        # Transpose the data points to have shape (N, p) [N points, p paths].
        data_points = data_points.T

        # Calculate distances between each data point and each concentration correction
        # point.
        # data_points has shape (N, p) [N points, p paths]
        # data_points[:, None, :] has shape (N, 1, p)
        # conc_corr_points has shape (M, p) [M points, p paths]
        # conc_corr_points[None, :, :] has shape (1, M, p) [M points]
        # The subtraction uses broadcasting to yield an array of shape (N, M, p),
        # np.linalg.norm(..., axis=2) computes the Euclidean distance resulting in an
        # array of shape [N, M].

        distances = np.linalg.norm(
            data_points[:, None, :] - conc_corr_points[None, :, :],
            axis=2,
        )

        threshold = np.finfo(np.float64).eps
        # For each data point, find the index where the distance is smaller
        # than the threshold. If the number of indices is different than 1, raise an
        # error.
        mask = distances < threshold
        if np.any(mask.sum(axis=1) != 1):
            raise ConcentrationCorrectionError(
                "There must be a single concentration correction point for each data "
                "point."
            )
        _, indices = np.where(mask)
        return indices.tolist()

    def create_conc_corr_scan(self, indices: list[int]) -> Scan:
        """
        Create a scan with the concentration correction data at the specified indices.

        Parameters
        ----------
        indices :
            Indices of the concentration correction data to be used.

        Returns
        -------
        Scan
            Scan with the concentration correction data at the specified indices.
        """
        signal = np.concatenate([scan.signal for scan in self.conc_corr_scans])
        scan = Scan(None, signal[indices])

        try:
            monitor = np.concatenate([scan.monitor for scan in self.conc_corr_scans])
            scan.monitor = monitor[indices]
        except IndexError:
            pass
        return scan

    def apply(self, scans: Scans) -> None:
        """
        Apply the concentration correction using data from the specified paths.

        Parameters
        ----------
        scans :
            Scans to be corrected.
        """
        logger.info("Applying data-informed concentration correction.")
        for scan in scans:
            indices = self.find_conc_corr_indices(scan)
            conc_corr_scan = self.create_conc_corr_scan(indices)
            scan.divide_by_scan(conc_corr_scan)


class DeadTimeCorrector(Corrector):
    """Class to perform dead time corrections."""


class Measurement:
    """Base class for measurements."""

    def __init__(self, sources: Source | list[Source]):
        """
        Parameters
        ----------
        sources :
            Sources of scans.
        """
        self.sources = [sources] if isinstance(sources, Source) else sources

        self._scans: Scans | None = None
        self._x: npt.NDArray[np.float64] = np.array([])
        self._signal: npt.NDArray[np.float64] = np.array([])
        self._monitor: npt.NDArray[np.float64] = np.array([])

    @property
    def scans(self) -> Scans:
        """The scans of the measurement."""
        if self._scans is None:
            self._scans = Scans()
            for source in self.sources:
                self._scans.extend(source.scans)

            if len(self._scans) == 0:
                raise ValueError("The measurement has no scans.")

            self._scans.check_sizes()

            self._x = np.array([])
            self._signal = np.array([])
            self._monitor = np.array([])
        return self._scans

    def add_source(self, source: Source) -> None:
        """
        Add a new source to the measurement.

        Parameters
        ----------
        source :
            Source to be added.
        """
        self.sources.append(source)
        self._scans = None

    def remove_source(self, index: int) -> None:
        """
        Remove a source from the measurement.

        Parameters
        ----------
        index :
            Index of the source to be removed.
        """
        self.sources.pop(index)
        self._scans = None

    def add_scans(self, scans: Scan | list[Scan]) -> None:
        """
        Add scans to the measurement.

        Parameters
        ----------
        scans :
            Scans to be added.
        """
        scans = [scans] if isinstance(scans, Scan) else scans

        if self._scans is None:
            self._scans = Scans()

        for scan in scans:
            self._scans.append(scan)
            logger.debug("Scan %s was added.", scan.label)

    def remove_scans(
        self, indices: int | list[int], filename: str | None = None
    ) -> None:
        """
        Remove scans from the measurement.

        Parameters
        ----------
        indices :
            Indices of the scans to be removed.
        filename :
            Name of the file from which the scans where read.
        """
        indices = [indices] if isinstance(indices, int) else indices

        for index in indices:
            for scan in self.scans:
                if index == scan.index and (
                    filename is None or scan.filename == filename
                ):
                    self.scans.remove(scan)
                    logger.debug("Scan %s was removed.", scan.label)

    def reset(self, scans: float = True):
        """Reset the measurement."""
        self._x, self._signal, self._monitor = np.array([]), np.array([]), np.array([])
        if scans:
            self.scans.reset()

    def get_scans(
        self, indices: int | list[int], filename: str | None = None
    ) -> list[Scan]:
        indices = [indices] if isinstance(indices, int) else indices

        scans = []
        for index in indices:
            for scan in self.scans:
                if scan.index == index and (
                    filename is None or scan.filename == filename
                ):
                    scans.append(scan)

        return scans

    # TODO: Should the method also accept an argument that limits the scans which
    # will be corrected?
    def concentration_correction(
        self,
        indices: int | list[int] | npt.NDArray[np.int64] | None = None,
        scans: Scan | list[Scan] | Scans | None = None,
        paths: dict[str, str] | None = None,
    ) -> None:
        """Apply the concentration correction using data from the specified scans.

        Parameters
        ----------
        indices :
            Indices of the scans used for concentration correction.
        scans :
            Scans used for concentration corrections.
        paths :
            Paths used to locate the required concentration correction data.
        """
        # Get the scans to be used for concentration correction.
        if indices is not None:
            indices = [indices] if isinstance(indices, int) else indices

            source = copy.deepcopy(self.sources[0])
            data_mappings = copy.deepcopy(source.data_mappings)

            with contextlib.suppress(KeyError):
                data_mappings["x"] = ".1/measurement/elapsed_time"
                logger.debug(
                    "The X-axis mapping for the concentration "
                    "correction scans was updated to elapsed time."
                )

            source.data_mappings = data_mappings
            scans = source.read_scans(indices)
        elif scans is not None:
            pass
        else:
            raise ValueError("Either the indices or scans must be specified.")

        conc_corr_scans = scans if isinstance(scans, Scans) else Scans(scans)

        if paths is None:
            corrector = SimpleConcentrationCorrector(conc_corr_scans)
        else:
            corrector = DataDrivenConcentrationCorrector(conc_corr_scans, paths)
        corrector.apply(self.scans)

        # Force signal and monitor reevaluation.
        self._signal, self._monitor = np.array([]), np.array([])


class Measurement1D(Measurement):
    """Base class for 1D measurements."""

    @property
    def x(self):
        if self._x.size == 0:
            self._x = self.scans.get_common_axis()
            # Assign the common axis to the scans.
            for scan in self.scans:
                scan.x = self._x
        return self._x

    @x.setter
    def x(self, a: npt.NDArray[np.float64]):
        logger.info("Setting new x-axis.")
        for scan in self.scans:
            scan.x = a
        self._x = a
        self._signal, self._monitor = np.array([]), np.array([])

    @property
    def signal(self):
        if self._signal.size == 0:
            self.process()
        return self._signal

    @property
    def monitor(self):
        if self._monitor.size == 0:
            self.process()
        return self._monitor

    def find_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Find outliers in the data.

        See the docstring of :meth:`.scans.Scan.find_outliers`.
        """
        for scan in self.scans:
            scan.find_outliers(method=method, **kwargs)

    def remove_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Remove outliers from the signal.

        See the docstring of :meth:`.scans.Scan.remove_outliers`.
        """
        logger.info("Removing outliers.")
        for scan in self.scans:
            scan.remove_outliers(method=method, **kwargs)
        self._signal = np.array([])

    def process(
        self, aggregation: str = "fraction of sums", normalization: str | None = None
    ):
        """
        Process the scans data.

        The processing includes aggregating the data of the selected scans
        and normalizing the signal.
        """
        self.aggregate(mode=aggregation)
        if normalization is not None:
            self.normalize(mode=normalization)

    def aggregate(self, mode: str = "fraction of sums"):
        """
        Aggregate the scans signal using the selected mode.

        When present, the aggregated monitor is always a sum of the monitors from
        the individual scans.

        Parameters
        ----------
        mode : str
            Defines how the signal is aggregated.

                - "sum" : Sum of the signals from all scans.
                - "fraction of sums" : Fraction of the signals sum and monitors sum.
                - "sum of fractions" : Sum of the signal and monitor fractions.

        """
        for scan in self.scans:
            if scan.monitor.size == 0:
                logger.info(
                    "No monitor data for scan %s. Setting aggregation mode to sum.",
                    scan.label,
                )
                mode = "sum"

        self._signal = np.zeros_like(self.x)
        if mode != "sum":
            self._monitor = np.zeros_like(self.x)

        for scan in self.scans:
            if mode in ("sum", "fraction of sums"):
                self._signal += np.nan_to_num(scan.signal, nan=0)
            elif mode in ("sum of fractions",):
                self._signal += np.nan_to_num(scan.signal, nan=0) / np.nan_to_num(
                    scan.monitor, nan=1
                )
            else:
                raise ValueError(f"Unknown aggregation mode {mode}.")

            # Aggregate monitor when mode is not "sum".
            if mode != "sum" and self._monitor.size > 0:
                self._monitor += np.nan_to_num(scan.monitor, nan=0)

        # Divide signal by monitor where monitor is not zero. If monitor is zero, the
        # signal is set to NaN.
        if mode == "fraction of sums":
            self._signal = np.where(
                self._monitor != 0, self._signal / self._monitor, np.nan
            )

        logger.info("The scans data was aggregated using the %s mode.", mode)

    def normalize(self, mode: str = "area") -> None:
        """
        Normalize the signal.

        Parameters
        ----------
        mode :
            Defines how the signal is normalized.

              - "area": Normalize using the absolute signal area calculated using the
                trapezoidal rule.
              - "maximum": Normalize using the absolute maximum intensity of the signal.

        Notes
        -----
        This will overwrite the original signal with the normalized one.
        """
        if self._signal.size == 0:
            self.aggregate()

        if mode == "area":
            self._signal = self._signal / np.abs(trapezoid(self._signal, self.x))
        elif mode == "maximum":
            self._signal = self._signal / np.abs(np.nanmax(self._signal))
        else:
            raise ValueError(f"Unknown normalization mode {mode}.")

        logger.info("The signal was normalized using the %s.", mode)

    def dead_time_correction(
        self,
        tau: Iterable[float],
        detection_time: float | npt.NDArray[np.float64] | None = None,
    ):
        """
        Perform a dead time correction using a non-paralyzable model.

        See the docstring of :meth:`.scans.Scan.dead_time_correction`.
        """
        for scan in self.scans:
            scan.dead_time_correction(tau, detection_time)

    def save(self, filename: str, delimiter: str = ","):
        """
        Save the current x and signal to file.

        Parameters
        ----------
        filename :
            Name of the output file.
        delimiter :
            Column delimiter in the output file.
        """
        if self.signal.size == 0:
            raise ValueError("The signal is not defined.")

        with open(filename, "w", encoding="utf-8") as fp:
            fp.write("# x signal\n")
            data = np.stack((self.x, self.signal), axis=1)
            np.savetxt(fp, data, delimiter=delimiter, fmt="%.6e %.6e")
            logger.info("The data was saved to %s.", filename)


class Xas(Measurement1D):
    """Class to represent a X-ray absorption measurement."""


class Xes(Measurement1D):
    """Class to represent a X-ray emission measurement."""


class Measurement2D(Measurement):
    """Base class for 2D measurements."""


class Rixs(Measurement2D):
    """Class to represent a resonant inelastic X-ray scattering measurement."""

    def __init__(self, sources: Source | list[Source]):
        super().__init__(sources=sources)
        self._y: npt.NDArray[np.float64] = np.array([])
        self._interpolator: Interpolator2D | None = None
        self.cuts = {}

    @property
    def x(self):
        if self._x.size == 0:
            self.process()
        return self._x

    @property
    def y(self):
        if self._y.size == 0:
            self.process()
        return self._y

    @property
    def signal(self):
        if self._signal.size == 0:
            self.process()
        return self._signal

    @property
    def interpolator(self):
        """The interpolator of the current plane."""
        if self._interpolator is None:
            self._interpolator = Interpolator2D(self.x, self.y, self.signal)
        return self._interpolator

    @property
    def acquisition_mode(self):
        """
        There are two ways to measure a RIXS plane:

        1. Step through a range of emission energies and scan the incoming
           (monochromator) energy for each step.
        2. Step through incoming (monochromator) energy and scan the emission energy.
        """
        if all(scan.y is not None and scan.y.size == 1 for scan in self.scans):
            mode = "absorption"
        else:
            mode = "emission"
        logger.debug("The RIXS plane was acquired in %s mode.", mode)
        return mode

    def reset(self, scans: float = True):
        """Reset the measurement."""
        super().reset(scans=scans)
        self._y = np.array([])
        self._interpolator = None

    def find_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Find outliers in the data.

        See the docstring of :meth:`.scans.Scan.find_outliers`.
        """
        for scan in self.scans:
            scan.find_outliers(method=method, **kwargs)

    def remove_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Remove outliers from the signal.

        See the docstring of :meth:`.scans.Scan.remove_outliers`.
        """
        logger.info("Removing outliers.")
        for scan in self.scans:
            scan.remove_outliers(method=method, **kwargs)
        self._signal = np.array([])

    def concentration_correction(
        self,
        indices: int | list[int] | None | npt.NDArray[np.int64] = None,
        scans: Scan | list[Scan] | Scans | None = None,
        paths: dict[str, str] | None = None,
    ) -> None:
        super().concentration_correction(indices, scans, paths)
        self._y = np.array([])
        self._interpolator = None

    def process(self):
        """Read and store the scans data."""
        acquisition_mode = self.acquisition_mode

        if acquisition_mode == "emission":
            raise NotImplementedError("The emission mode is not implemented yet.")

        x, y, signal = np.array([]), np.array([]), np.array([])

        if acquisition_mode == "absorption":
            for scan in self.scans:
                x = np.append(x, scan.x)
                y = np.append(y, scan.y * np.ones_like(scan.x))
                if scan.monitor.size == 0:
                    signal = np.append(signal, scan.signal)
                else:
                    signal = np.append(signal, scan.signal / scan.monitor)

            # Convert to energy transfer.
            y = x - y

        self._x, self._y, self._signal = x, y, signal

    def interpolate(self, xi: npt.NDArray[np.float64], yi: npt.NDArray[np.float64], /):
        """
        Interpolate the plane using the new axes.

        Parameters
        ----------
        xi :
            The new X-axis.
        yi :
            The new Y-axis.

        """
        x, y = np.meshgrid(xi, yi)
        x = xi.ravel()
        y = yi.ravel()
        points = np.stack((x, y), axis=-1)
        signal = self.interpolator(points)

        # Flatten arrays for storage.
        signal = signal.ravel()

        # Remove NaNs.
        mask = np.isfinite(signal)
        x = x[mask]
        y = y[mask]
        signal = signal[mask]

        # Assign the values.
        self._x, self._y, self._signal = x, y, signal

        # Update the interpolator.
        self.interpolator.update({"x": x, "y": y, "z": signal})

    def cut(
        self,
        mode: str = "CEE",
        energies: list[float] | None = None,
        npoints: int = 1024,
    ):
        """
        Calculate the cuts specified by the mode and energies.

        Parameters
        ----------
        mode : str
            Defines the way to cut the plane:

            - "CEE" - constant emission energy
            - "CIE" - constant incident energy
            - "CET" - constant energy transfer

        energies : list(float)
            Energies of the cuts.

        npoints : int
            Number of points for the cuts.

        """
        if energies is None:
            raise ValueError("The energies parameter must be defined.")

        mode = mode.upper()

        # Update the xc and yc arrays depending on the type of cut.
        for energy in energies:
            xc = np.linspace(self.x.min(), self.x.max(), npoints)
            yc = np.linspace(self.y.min(), self.y.max(), npoints)

            if mode == "CEE":
                yc = xc - np.ones_like(xc) * energy
            elif mode == "CIE":
                xc = np.ones_like(yc) * energy
            elif mode == "CET":
                yc = np.ones_like(xc) * energy
            else:
                raise ValueError(f"Unknown mode {mode}.")

            points = np.stack((xc, yc), axis=-1)
            signal = self.interpolator(points)

            if np.isnan(signal).all():
                logger.info("The %s cut at %s is empty.", mode, energy)
                continue

            # Remove NaNs.
            mask = np.isfinite(signal)
            xc = xc[mask]
            yc = yc[mask]
            signal = signal[mask]

            label = f"{mode.upper()}@{energy}"
            self.cuts[label] = (xc, yc, signal)



from silx.resources import ExternalResources

resources = ExternalResources(
    project="daxs",
    url_base="https://spectroscopy.gitlab-pages.esrf.fr/daxs-data",
    env_key="DAXS_DATA",
    timeout=60,
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from daxs.scans import Scan, Scans

import numpy as np
import pytest

from daxs.filters import hampel
from daxs.sources import Hdf5Source
from daxs.utils import resources


@pytest.fixture
def scans() -> Scans:
    filename = resources.getfile("Pd_foil_La_XANES.h5")
    data_mappings = {
        "x": ".1/measurement/hdh_angle",
        "signal": [".1/measurement/g09", ".1/measurement/g14"],
    }
    source = Hdf5Source(filename, 5, data_mappings=data_mappings)
    return source.scans


@pytest.mark.parametrize("indices, values", [(69, 14111.5)])
def test_hampel(scans: list[Scan], indices: int, values: float):
    [scan] = scans
    outliers, medians = hampel(scan.signal, window_size=5, threshold=3.5, axis=0)
    assert outliers[indices]
    assert medians[indices] == pytest.approx(values, abs=np.finfo(float).eps)

    outliers, medians = hampel(scan.signal, window_size=20, threshold=3.5, axis=0)
    assert not outliers[indices]



from __future__ import annotations

import copy
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pytest

from daxs.measurements import Measurement
from daxs.scans import Scan
from daxs.sources import Hdf5Source
from daxs.utils import resources


@pytest.fixture
def scan():
    x = np.array([3, 1, 2, 0, 4], dtype=float)
    signal = np.array([[2, 9, 0, 4, 1], [9, 1, 3, 4, 3]], dtype=float)
    data = {
        "monitor": np.array([1, 1, 2, 4, 2], dtype=float),
        "detection_time": [0.2, 0.2, 0.2, 0.2, 0.2],
        "filename": "No file name",
        "index": 1,
    }
    scan = Scan(x, signal, data=data)
    return scan


def test_scan_init():
    rng = np.random.default_rng()
    x = rng.random(10)
    signal = "not an array"

    with pytest.raises(TypeError):
        Scan(x, signal)  # type: ignore


def test_scan_reset(scan: Scan):
    scan.x = np.array([1, 1, 1, 1, 1])
    scan.reset()
    assert scan.x == pytest.approx([0, 1, 2, 3, 4])


def test_scan_properties(scan: Scan):
    assert scan.x == pytest.approx([0, 1, 2, 3, 4])
    assert scan.signal == pytest.approx([4.0, 5.0, 1.5, 5.5, 2.0])
    assert scan.monitor == pytest.approx([4, 1, 2, 1, 2])

    scan.x = np.array([3, 1, 2, 0, 4])
    assert scan.x == pytest.approx([0, 1, 2, 3, 4])
    assert scan.signal == pytest.approx([4.0, 5.0, 1.5, 5.5, 2.0])

    scan.reset()
    scan.x = np.array([5, 6, 7, 9, 11])
    assert scan.x == pytest.approx([5, 6, 7, 9, 11])
    assert scan.signal == pytest.approx([4.0, 5.0, 1.5, 5.5, 2.0])

    scan.reset()
    with pytest.raises(ValueError):
        scan.x = np.array([5, 6, 7, 9, 11, 12])

    scan.reset()
    scan.x = np.array([0, 1, 2, 3, 5, 10])
    assert scan.x == pytest.approx([0, 1, 2, 3, 5, 10])
    assert scan.signal == pytest.approx(
        [4.0, 5.0, 1.5, 5.5, np.nan, np.nan], nan_ok=True
    )


def test_scan_interpolate(scan: Scan):
    scan._signal = np.array([])  # type: ignore
    with pytest.raises(ValueError):
        rng = np.random.default_rng()
        scan.interpolate(a=rng.random(10))


def test_scan_outliers_removal(scan: Scan):
    scan.remove_outliers(method="hampel")
    assert scan.signal == pytest.approx([4.0, 5.0, 1.5, 2.5, 2.0])


def test_scan_dead_time_correction(scan: Scan):
    with pytest.raises(TypeError):
        scan.dead_time_correction()  # type: ignore

    tau = np.array([1.0, 1.0, 1.0], dtype=np.float64) * 1e-3
    with pytest.raises(ValueError):
        scan.dead_time_correction(tau=tau)

    scan.reset()
    tau = np.array([1.0, 1.0]) * 1e-3
    scan.dead_time_correction(tau)
    assert scan.signal == pytest.approx(
        np.array([4.08163265, 5.21455445, 1.52284264, 5.72214289, 2.0253552])
    )

    scan.reset()
    scan.data.pop("detection_time", None)
    with pytest.raises(ValueError):
        scan.dead_time_correction(tau)

    with pytest.raises(ValueError):
        scan.dead_time_correction(tau=tau, detection_time=0.0)

    with pytest.raises(ValueError):
        scan.dead_time_correction(tau=[1.0, 1.0], detection_time=2)


def test_scan_plot(scan: Scan):
    _, ax = plt.subplots()
    scan.remove_outliers(method="hampel")
    scan.plot(ax)


@pytest.mark.filterwarnings("ignore::RuntimeWarning")
def test_scan_divide_by_scan(scan: Scan):
    scan1 = copy.deepcopy(scan)
    scan2 = copy.deepcopy(scan)

    scan2.divide_by_scan(scan1)
    scan2._signal = scan2._signal / 2.0  # type: ignore
    assert scan2.signal[0:2] == pytest.approx([0.5, 0.5])


def test_str(scan: Scan):
    assert str(scan) == "None/None"


@pytest.fixture()
def hdf5_filename():
    return resources.getfile("Pd_foil_La_XANES.h5")


@pytest.fixture()
def data_mappings():
    return {
        "x": ".1/measurement/hdh_angle",
        "signal": [".1/measurement/g09", ".1/measurement/g14"],
    }


def test_scans_get_common_axis(hdf5_filename: str, data_mappings: dict[str, Any]):
    source = Hdf5Source(hdf5_filename, [3], data_mappings=data_mappings)
    measurement = Measurement(source)
    values = measurement.scans.get_common_axis("x")
    assert np.all(values == getattr(measurement.scans[0], "x"))

    source = Hdf5Source(hdf5_filename, [3, 4, 7, 8, 9], data_mappings=data_mappings)
    measurement = Measurement(source)

    values = measurement.scans.get_common_axis("x")
    assert values[-1] == pytest.approx(38.72936736)

    values = measurement.scans.get_common_axis("x", mode="union")
    assert values[-1] == pytest.approx(38.72939236)



"""The module provides functions for filtering data."""

from __future__ import annotations

import logging

import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)


def hampel(
    data: npt.NDArray[np.float64],
    window_size: int | None = None,
    threshold: float = 3.5,
    axis: int = 0,
    k: float = 1.4826,
):
    """
    Outliers detection using the Hampel filter.

    More details about the filter can be found here:

    - https://fr.mathworks.com/help/dsp/ref/hampelfilter.html
    - https://dsp.stackexchange.com/questions/26552
    - https://stackoverflow.com/questions/22354094

    Parameters
    ----------
    data :
        Input data.
    window_size :
        Size of the sliding window.
    threshold :
        Threshold for outlier detection expressed in number of standard
        deviations. Iglewicz and Hoaglin [1]_ suggest using a value of 3.5, but
        larger values are often needed to avoid removing data points from a
        noisy signal.
    axis :
        Axis along which the detection is performed.
    k :
        Scale factor for the median absolute deviation. The default value is
        1.4826, which is the scale factor for normally distributed data.

    Returns
    -------
    np.NDArray[np.bool_]
        Mask identifying the outliers, and the rolling window median.

    References
    ----------
    .. [1] Boris Iglewicz and David Hoaglin (1993) "Volume 16: How to Detect
       and Handle Outliers", The ASQC Basic References in Quality Control:
       Statistical Techniques, Edward F. Mykytka, Ph.D., Editor.

    """
    if window_size is None:
        SECTIONS: int = 10
        window_size = data.shape[axis] // SECTIONS

    # Don't allow the window size to be too small or larger than the data.
    MIN_WINDOW_SIZE = 3
    window_size = max(window_size, MIN_WINDOW_SIZE)
    window_size = min(window_size, data.shape[axis])

    # Make the size of the window an odd number.
    if window_size % 2 == 0:
        window_size -= 1

    logger.debug("Hampel filter window size = %d.", window_size)

    # Pad the data along the direction used for the outlier detection.
    pad_size = np.zeros((len(data.shape), 2), dtype=int)
    pad_size[axis] = np.array((window_size // 2, window_size // 2))
    padded_data = np.pad(data, pad_size)

    # Create a sliding window view of the padded data.
    padded_data_views = np.lib.stride_tricks.sliding_window_view(
        padded_data, window_size, axis, subok=True
    )

    # Compute the median of each view. As the views increase the number of dimensions,
    # the axis argument always identifies the axis along which the median is computed.
    medians = np.median(padded_data_views, axis=-1)
    abs_diff = np.abs(data - medians)

    outliers = np.full(data.shape, False, dtype=bool)

    # Put the median in a suitable shape.
    medians_views = np.repeat(medians, window_size).reshape(padded_data_views.shape)

    # Calculate the median absolute deviation for each view.
    mad = np.median(np.abs(padded_data_views - medians_views), axis=-1)

    # Calculate the standard deviation. We assume that the data is normally distributed.
    # https://en.wikipedia.org/wiki/Median_absolute_deviation
    sigma = k * mad

    score = threshold * sigma
    outliers[(abs_diff > score)] = True

    # If the score is zero, the data is not necessarily an outlier because all
    # absolute differences with the exception of zero will be greater than it.
    # For this case we add a condition that excludes from the outliers the points
    # that are smaller than the median.
    outliers[(score == 0) & (np.abs(data) < np.abs(np.median(data)))] = False

    return outliers, medians



"""The module provides classes for the representation of scans in measurements."""

from __future__ import annotations

import copy
import logging
from typing import TYPE_CHECKING, Any, Iterable

if TYPE_CHECKING:
    import matplotlib.axes
import numpy as np
import numpy.typing as npt

from daxs.config import Config
from daxs.filters import hampel
from daxs.utils import arrays

logger = logging.getLogger(__name__)

use_blissdata_api = Config().get("use_blissdata_api", False)
if use_blissdata_api:
    from blissdata.h5api.dynamic_hdf5 import File
else:
    from silx.io.h5py_utils import File


class Scan:
    def __init__(
        self,
        x: npt.NDArray[np.float64] | None = None,
        signal: npt.NDArray[np.float64] | None = None,
        *,
        data: dict[Any, Any] | None = None,
    ) -> None:
        """Define the base representation of scans in measurements.

        An scan can be initialized in two ways:

        1. Direct initialization - by passing the `x` and `signal` directly to
           the constructor, with optional additional data.

        2. Lazy initialization - by creating an empty scan and then setting values
           through properties, either all at once via the `data` property or
           individually `x`, `signal`, `monitor`, etc.

        Parameters
        ----------
        x :
            X-axis values (1D array).
        signal :
            Signal values (1D or 2D array). For a 2D array, the components must be
            stored as rows. A 1D array will be converted to a 2D array.
        data :
            Storage for the raw scan data and metadata.
        """
        self._data = {} if data is None else data

        # Initialize empty arrays.
        self._x = np.array([])
        self._y = np.array([])
        self._signal = np.array([])
        self._monitor = np.array([])

        # Array of indices used to reindex the data.
        self._indices: npt.NDArray[np.int32] = np.array([])

        # Initialize the outliers and medians arrays.
        self.outliers: npt.NDArray[np.bool_] = np.array([])
        self.medians: npt.NDArray[np.float64] = np.array([])

        self.filename: str | None = None
        self.index: int | None = None

        self.aggregation: str = "mean"

        # Set the data from the input parameters.
        self.x = x if x is not None else np.array([])
        self.signal = signal if signal is not None else np.array([])
        for attr in ("y", "monitor"):
            if attr in self._data:
                setattr(self, attr, self._data[attr])
        # Reindex the data using the indices array that sorts the X-axis values.
        self.reindex()

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, a: npt.NDArray[np.float64]) -> None:
        """Set the X-axis values.

        Several cases are considered:

        1. The first assignment. In this case, the new values are simply stored.
        2. The new values are the same as the current ones. In this case, nothing has
           to be done.
        3. The limits of the new values are within the current values. In this case,
           the signal and monitor data are interpolated to the new X-axis values.
        4. The new values are outside the current values, but the two arrays have the
           same shape. In this case, the new values are assigned to the X-axis. It
           is useful when the X-axis changes to different units, e.g., angle to energy,
           which does not require interpolation.
        5. The new values are outside the current values and of different shapes. In
           this case, an error is raised.

        The x values are used as the independent variable for all scan data. When
        changing x values, dependent variables like signal and monitor may need to be
        interpolated to maintain correct data relationships.

        Parameters
        ----------
        a : npt.NDArray[np.float64]
            The new x-axis values to set.

        Returns
        -------
        None

        Raises
        ------
        TypeError
            If the input is not a NumPy array.
        ValueError
            If the new values are outside the range of the current values and
            the arrays have different shapes.
        """
        if not isinstance(a, np.ndarray):
            raise TypeError("The X-axis values must be a Numpy array.")

        # First assignment case.
        if "x" not in self._data:
            self._data["x"] = a
            self._x = a.astype(np.float64)
            return

        # Normal case with existing data.
        a = np.sort(a, kind="stable")

        if np.array_equal(self._x, a):
            logger.debug("The new X-axis values are the same as the current ones.")
            return
        elif arrays.intersect(a, self._x).size > 0:
            self.interpolate(a)
            return
        elif self._x.size == a.size:
            logger.debug("Assigning the new X-axis values.")
            self._x = a.astype(np.float64)
            self._indices = np.array([])
            return
        else:
            raise ValueError("The new X-axis values are outside the current ones.")

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, a: npt.NDArray[np.float64]) -> None:
        """Set the Y-axis values.

        Parameters
        ----------
        a :
            Y-axis values (1D array).
        """
        if not isinstance(a, np.ndarray):
            raise TypeError("The Y-axis values must be a Numpy array.")

        if "y" not in self._data:
            self._data["y"] = a

        self._y = a.astype(np.float64)

    @property
    def signal(self):
        if self.aggregation == "mean":
            return self._signal.mean(axis=0)
        else:
            raise ValueError(f"Unknown aggregation method: {self.aggregation}.")

    @signal.setter
    def signal(self, a: npt.NDArray[np.float64]) -> None:
        """Set the signal values.

        Parameters
        ----------
        a :
            Signal values (1D or 2D array). For a 2D array, the components must be
            stored as rows. A 1D array will be converted to a 2D array.
        """
        if not isinstance(a, np.ndarray):
            raise TypeError("The signal values must be a Numpy array.")

        if a.ndim not in (1, 2):
            raise ValueError("The signal must be a 1D or a 2D array.")

        # Check if signal matches current x dimension.
        if self._x.size > 0 and a.size > 0 and self._x.size != a.shape[-1]:
            raise ValueError(
                f"The signal size ({a.shape[-1]}) must match the "
                f"X-axis size ({self._x.size})."
            )

        if "signal" not in self._data:
            self._data["signal"] = a

        self._signal = a.astype(np.float64)

        # Convert 1D signal to 2D for consistent internal representation.
        if self._signal.ndim == 1:
            self._signal = self._signal[np.newaxis, :]

    @property
    def monitor(self):
        return self._monitor

    @monitor.setter
    def monitor(self, a: npt.NDArray[np.float64]) -> None:
        """Set the monitor values.

        Parameters
        ----------
        a :
            Monitor values (1D array).
        """
        if not isinstance(a, np.ndarray):
            raise TypeError("The monitor values must be a Numpy array.")

        # Check if monitor size matches current x dimension.
        if self._x.size > 0 and a.size > 0 and a.size != self._x.size:
            raise ValueError(
                f"The monitor size ({a.size}) must match the "
                "X-axis size ({self._x.size})."
            )

        if "monitor" not in self._data:
            self._data["monitor"] = a

        self._monitor = a.astype(np.float64)

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, a: npt.NDArray[np.int32]) -> None:
        """Set the indices array.

        Parameters
        ----------
        a :
            Indices array (1D array).
        """
        if not isinstance(a, np.ndarray):
            raise TypeError("The indices must be a Numpy array.")

        if a.shape != self._x.shape:
            raise ValueError("The indices and X-axis arrays must have the same shape.")

        self._indices = a.astype(np.int32)
        self.reindex()

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data: dict[Any, Any]) -> None:
        """Set the data dictionary."""
        if not isinstance(data, dict):
            raise TypeError("The data must be a dictionary.")
        self._data = data
        self.reset()

    @property
    def label(self) -> str:
        return f"{self.filename}/{self.index}"

    def reset(self) -> None:
        """Reset the scan data to the original values stored internally."""
        try:
            self._x = self._data["x"].astype(np.float64)
        except KeyError as e:
            raise KeyError("The data dictionary does not contain X-axis values.") from e

        try:
            self._y = self._data["y"].astype(np.float64)
        except KeyError:
            logger.debug("The data dictionary does not contain Y-axis values.")

        try:
            self._signal = self._data["signal"].astype(np.float64)
        except KeyError as e:
            raise KeyError("The data dictionary does not contain signal values.") from e

        # The stored signal may be a 1D array, so we need to convert it to a 2D array.
        if self._signal.ndim == 1:
            self._signal = self._signal[np.newaxis, :]

        try:
            self._monitor = self._data["monitor"].astype(np.float64)
        except KeyError:
            logger.debug("The data dictionary does not contain monitor values.")

        self._indices = np.array([])
        self.outliers, self.medians = np.array([]), np.array([])

        # Only reindex if we have valid data.
        if self._x.size > 0 and self._signal.size > 0:
            self.reindex()
        else:
            logger.warning(
                f"The x and signal arrays are empty for scan {self.label}. "
                "Skipping reindexing."
            )

    def reindex(self):
        """Reindex the scan data."""
        if self._x.size == 0:
            return
        if self._indices.size == 0:
            self._indices = np.argsort(self._x, kind="stable")
        self._x = self._x[self._indices]
        self._signal = self._signal[:, self._indices]
        if self._monitor.size != 0:
            self._monitor = self._monitor[self._indices]

    def read_data_at_paths(
        self, data_paths: str | list[str]
    ) -> npt.NDArray[np.float64]:
        """Read and store the data from the file."""
        if self.filename is None:
            raise ValueError("The filename from where to read the data must be set.")
        if self.index is None:
            raise ValueError("The scan index from where to read the data must be set.")

        data_paths = [data_paths] if isinstance(data_paths, str) else data_paths

        kwargs: dict[Any, Any] = {}
        if use_blissdata_api:
            kwargs["retry_timeout"] = Config().get("dynamic_hdf5_retry_timeout")

        data: list[Any] = []
        with File(self.filename, **kwargs) as fh:
            for data_path in data_paths:
                full_data_path = f"{self.index}{data_path}"

                try:
                    data_at_path = fh[full_data_path][()]  # type: ignore
                except KeyError as e:
                    raise KeyError(f"Unable to access {full_data_path}.") from e
                except TypeError as e:
                    raise TypeError(
                        f"Unable to read data from {full_data_path}."
                    ) from e

                try:
                    data_at_path = np.asarray(data_at_path)  # type: ignore
                except ValueError as e:
                    raise ValueError(
                        f"Unable to convert data from {full_data_path} "
                        "to a Numpy array."
                    ) from e

                if data_at_path.size == 0:
                    raise ValueError(f"Data from {full_data_path} is empty.")

                data.append(data_at_path)

        # Return the element of the array if it has only one element.
        if len(data) == 1:
            [data] = data

        return np.array(data)

    def find_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Find outliers in the signal.

        See the docstring in the :mod:`daxs.filters`.
        """
        if method == "hampel":
            self.outliers, self.medians = hampel(self._signal, axis=1, **kwargs)
        else:
            raise ValueError(f"Unknown method: {method}.")

    def remove_outliers(self, method: str = "hampel", **kwargs: Any):
        """
        Remove outliers from the signal.

        See the docstring of :meth:`daxs.scans.Scan.find_outliers`.
        """
        if self.outliers.size == 0 or self.medians.size == 0:
            self.find_outliers(method=method, **kwargs)

        if self.outliers.size > 0 and self.medians.size > 0:
            self._signal = np.where(self.outliers, self.medians, self._signal)
        else:
            logger.info("No outliers found for scan %s.", self.label)

    def dead_time_correction(
        self,
        tau: Iterable[float],
        detection_time: float | npt.NDArray[np.float64] | None = None,
    ):
        """
        Perform a dead time correction using a non-paralyzable model.

        Parameters
        ----------
        tau :
            The detector dead time in seconds.
        detection_time :
            The time spent on a point of the scan in seconds.

        """
        if detection_time is None:
            try:
                detection_time = copy.deepcopy(self._data["detection_time"])
            except KeyError:
                raise ValueError(
                    "Either the detection time parameter or `detection_time`"
                    " data path must be set."
                )
        else:
            detection_time = np.ones_like(self.signal) * detection_time

        detection_time = np.asarray(detection_time)

        if np.any(detection_time == 0):
            raise ValueError("The detection time has zero values.")

        tau = np.array(tau)
        if self._signal.shape[0] != tau.shape[0]:
            raise ValueError(
                "Each signal data path must have a detector dead time (tau) value."
            )

        norm = 1 - ((self._signal / detection_time).T * tau).T
        if np.any(norm == 0):
            raise ValueError("The normalization has zero values.")

        self._signal = self._signal / norm

    # TODO: Extract the interpolation logic to a separate class.
    def interpolate(self, a: npt.NDArray[np.float64]):
        """
        Interpolate the signal and possibly the monitor data to the new X-axis values.

        Parameters
        ----------
        a :
            Array used to interpolate the signal and monitor.

        """
        if a.size == 0:
            raise ValueError("The new X-axis values can not be empty.")

        if self._signal.size == 0:
            raise ValueError("The signal values must not be empty.")

        logger.debug(
            "Interpolating the signal and monitor data for scan %s.", self.label
        )

        # The interpolated signal is probably going to have a different size,
        # so we can't change the values in-place, and a new array needs to be
        # initialized.
        signal = np.zeros((self._signal.shape[0], a.size))

        # Interpolate the signal from each counter individually.
        for i, _ in enumerate(self._signal):
            signal[i, :] = np.interp(
                a, self._x, self._signal[i, :], left=np.nan, right=np.nan
            )

        # Interpolate the monitor if present.
        if self._monitor.size > 0:
            self._monitor = np.interp(
                a, self._x, self._monitor, left=np.nan, right=np.nan
            )

        self._x = a
        self._signal = signal
        self._indices = np.array([])

    def divide_by_scalars(
        self, signal: int | float, monitor: int | float | None = None
    ) -> Scan:
        """Divide the scan by scalar values."""
        self._signal /= signal
        if monitor is not None:
            self._monitor /= monitor
        return self

    def divide_by_scan(self, other: Scan) -> Scan:
        """Divide the scan by another scan."""
        if not isinstance(other, Scan):
            raise TypeError("The divisor must be a scan.")

        try:
            self._signal /= np.nan_to_num(other._signal, nan=1)
        except ValueError as e:
            raise ValueError(
                "The signal arrays of the two scans must have the same shape."
            ) from e

        # Check for empty or all-zero divisors.
        if other._signal.size == 0:
            raise ValueError("Cannot divide an empty signal.")
        if np.all(other._signal == 0):
            raise ValueError("Cannot divide by signal with all zero values.")

        if self._monitor.size > 0 and other._monitor.size > 0:
            if np.all(other._monitor == 0):
                logger.warning(
                    "Monitor values are all zero. This may result in invalid data."
                )
            try:
                self._monitor /= np.nan_to_num(other._monitor, nan=1)
            except ValueError as e:
                raise ValueError(
                    "The monitor arrays of the two scans must have the same shape."
                ) from e

        return self

    def __truediv__(self, other: Scan) -> Scan:
        """Divide the scan by another scan."""
        return self.divide_by_scan(other)

    def plot(self, ax: matplotlib.axes.Axes, shift: float = 0.0):
        """
        Plot the scan data and outliers if available.

        Parameters
        ----------
        ax :
            The axes to plot the scan data on.
        shift :
            Shift the signal by the given value.

        """
        shift = float(np.mean(self._signal))
        for i, _ in enumerate(self._signal):
            ax.plot(self.x, self._signal[i, :] + i * shift, label=f"{i}")
            if self.outliers.size > 0:
                indices = self.outliers[i, :]
                ax.plot(self.x[indices], self._signal[i, :][indices] + i * shift, "k.")
            ax.legend()

    def __str__(self):
        return self.label


class Scans:
    """A collection of scans."""

    def __init__(self, scans: Scan | list[Scan] | None = None) -> None:
        """Initialize the collection of scans."""
        if scans is None:
            self.scans = []
        elif isinstance(scans, list):
            self.scans = scans
        else:
            self.scans = [scans]

    def check_sizes(self) -> None:
        """Sanity check for the number of points in the scans."""
        sizes = [scan.x.size for scan in self.scans]
        mean = np.mean(sizes)
        std = np.std(sizes)

        if any(abs(size - mean) > std for size in sizes):
            logger.warning(
                "The number of points in the selected scans have a "
                "large spread (mean = %.2f, standard deviation: %.2f).",
                mean,
                std,
            )

    # TODO: Extract the common axis logic to a utility function.
    def get_common_axis(
        self, label: str = "x", mode: str = "intersection"
    ) -> npt.NDArray[np.float64]:
        """Return the common axis for the scans."""
        if not self.scans:
            raise ValueError("There are no scans available.")

        def step(axis: npt.NDArray[np.float64]) -> float:
            return np.abs((axis[0] - axis[-1]) / (axis.size - 1))

        # If there is a single scan, use its axis as the common axis.
        if len(self.scans) == 1:
            [axis] = self.scans
            return getattr(axis, label)

        axes = sorted([getattr(scan, label) for scan in self.scans], key=np.min)

        # Initialize the common axis as the first axis.
        common_axis = axes[0]
        for i, axis in enumerate(axes):
            message = (
                f"{label.upper()}-axis parameters for scan {self.scans[i].label}: "
                f"start = {axis[0]:.8f}, stop = {axis[-1]:.8f}, "
                f"size = {axis.size:d}, step = {step(axis):.8f}."
            )
            logger.debug(message)

            if np.array_equal(common_axis, axis):
                continue

            common_axis = arrays.merge(common_axis, axis, mode=mode)

            if common_axis.size == 0 and mode == "intersection":
                message = (
                    f"The common {label.upper()}-axis is empty after merging scan "
                    f"{self.scans[i].label}. "
                    "Switching to union mode for the common axis search."
                )
                logger.warning(message)
                return self.get_common_axis(label, mode="union")

        message = (
            f"Common {label.upper()}-axis parameters using {mode} mode: "
            f"start = {common_axis[0]:.8f}, stop = {common_axis[-1]:.8f}, "
            f"size = {common_axis.size:d}, step = {step(common_axis):.8f}"
        )
        logger.info(message)

        return common_axis

    def reset(self) -> None:
        """Reset the scans to their original values."""
        for scan in self.scans:
            scan.reset()

    def extend(self, scans: Scans) -> None:
        """Extend the collection of scans."""
        self.scans.extend(scans)

    def __len__(self) -> int:
        """Return the number of scans in the collection."""
        return len(self.scans)

    def __iter__(self):
        """Iterate over the scans."""
        return iter(self.scans)

    def __getitem__(self, index: int) -> Scan:
        """Return the scan at the given index."""
        return self.scans[index]

    def __str__(self):
        """Return a string representation of the collection."""
        return "\n".join([str(scan) for scan in self.scans])

    def remove(self, item: Scan) -> None:
        """Remove the scan at the given index."""
        self.scans.remove(item)

    def append(self, item: Scan) -> None:
        """Append a scan to the collection."""
        self.scans.append(item)



"""Module containing array utilities."""

from __future__ import annotations

import logging
from typing import Any

import numpy as np
import numpy.typing as npt

logger = logging.getLogger(__name__)


def intersect(
    a: npt.NDArray[np.float64], b: npt.NDArray[np.float64], sort: bool = False
) -> npt.NDArray[np.float64]:
    """Compute the intersection of two arrays.

    The function sorts the input arrays if required, determines the indices of
    the intersecting region, and returns the intersection. If there is no
    intersection, an empty array is returned.

    Parameters
    ----------
    a :
        The first input array.
    b :
        The second input array.
    sort :
        If True, the input arrays are sorted; default is False.

    Returns
    -------
    intersection :
        The intersection of the input arrays. If there is no intersection, an
        empty array is returned.
    """
    # Sort the input arrays if required.
    if sort:
        a, b = np.sort(a, kind="stable"), np.sort(b, kind="stable")

    # Sort the input arrays by their minimum value.
    a, b = sorted([a, b], key=np.min)

    # Determine the indices of the intersecting region.
    start, stop = np.max([a[0], b[0]]), np.min([a[-1], b[-1]])
    ida = (a >= start) & (a <= stop)
    idb = (b >= start) & (b <= stop)

    if not len(a[ida]) and not len(b[idb]):
        return np.array([])
    elif len(a[ida]) == 1 and len(b[idb]) == 1:
        return a[ida]
    elif len(b[idb]) == 1:
        intersection = a[ida]
    elif len(a[ida]) == 1:
        intersection = b[idb]
    elif np.abs(np.mean(np.diff(a[ida]))) <= np.abs(np.mean(np.diff(b[idb]))):
        intersection = a[ida]
    else:
        intersection = b[idb]

    # Add start and stop points if they are not included in the intersection.
    if start not in intersection:
        intersection = np.concatenate([[start], intersection])
    if stop not in intersection:
        intersection = np.concatenate([intersection, [stop]])

    return intersection


def union(
    a: npt.NDArray[np.float64], b: npt.NDArray[np.float64], sort: bool = False
) -> npt.NDArray[np.float64]:
    # Sort the input arrays if required.
    if sort:
        a, b = np.sort(a), np.sort(b)

    # Sort the input arrays by their minimum value.
    a, b = sorted([a, b], key=np.min)

    # Determine the indices of the overlapping regions.
    start, stop = np.max([a[0], b[0]]), np.min([a[-1], b[-1]])

    intersection = intersect(a, b, sort=False)

    # The if statement is always ran because the arrays are already sorted by their
    # minimum value.
    if a[0] < b[0]:
        union = np.concatenate([a[a < start], intersection])
    else:
        union = np.concatenate([b[b < start], intersection])

    # Check the last element of the arrays to decide which array append.
    if a[-1] > b[-1]:
        union = np.concatenate([union, a[a > stop]])
    else:
        union = np.concatenate([union, b[b > stop]])

    return union


def merge(
    a: npt.NDArray[Any],
    b: npt.NDArray[Any],
    mode: str = "intersection",
    sort: bool = False,
) -> npt.NDArray[Any]:
    """Merge two array using the specified mode."""
    if mode == "intersection":
        merged = intersect(a, b, sort)
    elif mode == "union":
        merged = union(a, b, sort)
    else:
        raise ValueError("Invalid mode")
    return merged


def discretize_intervals(intervals: list[list[float]]) -> npt.NDArray[np.float64]:
    # Check if the start energy is smaller than the end energy for all regions.
    if not all(interval[0] < interval[1] for interval in intervals):
        raise ValueError(
            "The start energy must be smaller than the end energy for all regions."
        )

    # Check if the step is positive for all regions.
    if not all(interval[2] > 0 for interval in intervals):
        raise ValueError("The step must be positive for all regions.")

    # Check if the start energies of the regions are in increasing order.
    if not all(
        intervals[i][0] <= intervals[i + 1][0] for i in range(len(intervals) - 1)
    ):
        raise ValueError(
            "The start energies of the regions must be in increasing order."
        )

    points = []
    for i, (start, end, step) in enumerate(intervals):
        points.extend(np.arange(start, end, step))
        # Add the end point of the last region or if it does not intersection with
        # the next region.
        if i == len(intervals) - 1 or end < intervals[i + 1][0]:
            points.append(end)

    return np.array(points)


def trapezoid(
    y: npt.NDArray[np.float64], x: npt.NDArray[np.float64] | None = None
) -> float:
    """
    Integrate `y` using the composite trapezoidal rule while ignoring NaN values.

    Parameters
    ----------
    y :
        The values of the function to integrate. NaN values are ignored.
    x :
        The sample points corresponding to `y` values. If None, points are
        assumed to be spaced at a unit distance.

    Returns
    -------
    float
        The estimated integral of `y` using the trapezoidal rule.
    """
    y = np.asarray(y)
    x = np.arange(len(y)) if x is None else np.asarray(x)

    # Ensure x and y have the same shape
    if y.shape != x.shape:
        raise ValueError("Shape of x and y must be the same.")

    # Identify non-NaN regions
    mask = ~np.isnan(y)
    y = y[mask]
    x = x[mask]

    # Perform trapezoidal integration on non-NaN regions
    if y.size < 2:  # noqa: PLR2004
        return 0  # Not enough points to perform integration

    if np.version.version < "2.0.0":
        return np.trapz(y, x)  # noqa: NPY201
    else:
        return np.trapezoid(y, x)  # type: ignore



"""Material class and related functions."""

from __future__ import annotations

from typing import Iterable

import numpy as np

MATERIALS = {
    "Si": [5.43071, 5.43071, 5.43071, 90.0, 90.0, 90.0],
    "Ge": [5.658, 5.658, 5.658, 90.0, 90.0, 90.0],
    "LiNbO3": [5.1501, 5.1501, 5.4952, 62.057, 62.057, 60.0],
    "a-quartz": [4.9133, 4.9133, 5.4053, 90.0, 90.0, 120.0],
    "diamond": [3.567, 3.567, 3.567, 90.0, 90.0, 90.0],
}


class Material:
    def __init__(  # noqa: PLR0913
        self,
        a: float,
        b: float,
        c: float,
        alpha: float,
        beta: float,
        gamma: float,
        name: str | None = None,
    ) -> None:
        """Initialize a material with its unit cell parameters."""
        self.a = a
        self.b = b
        self.c = c
        self.alpha = np.radians(alpha)
        self.beta = np.radians(beta)
        self.gamma = np.radians(gamma)
        self.name = name

    # fmt: off
    @property
    def volume(self) -> float:
        return (
            self.a * self.b * self.c * np.sqrt(
                1
                - np.cos(self.alpha) ** 2
                - np.cos(self.beta) ** 2
                - np.cos(self.gamma) ** 2
                + 2 * np.cos(self.alpha) * np.cos(self.beta) * np.cos(self.gamma)
            )
        )

    @property
    def s11(self) -> float:
        return self.b**2 * self.c**2 * np.sin(self.alpha) ** 2

    @property
    def s22(self) -> float:
        return self.a**2 * self.c**2 * np.sin(self.beta) ** 2

    @property
    def s33(self) -> float:
        return self.a**2 * self.b**2 * np.sin(self.gamma) ** 2

    @property
    def s12(self) -> float:
        return (
            self.a * self.b * self.c**2
            * (np.cos(self.alpha) * np.cos(self.beta) - np.cos(self.gamma))
        )

    @property
    def s23(self) -> float:
        return (
            self.a**2 * self.b * self.c
            * (np.cos(self.beta) * np.cos(self.gamma) - np.cos(self.alpha))
        )

    @property
    def s31(self) -> float:
        return (
            self.a * self.b**2 * self.c
            * (np.cos(self.gamma) * np.cos(self.alpha) - np.cos(self.beta))
        )
    # fmt: on

    def calculate_d_spacing(self, reflection: Iterable[int] = (1, 1, 1)):
        """Calculate the d spacing for a given reflection."""
        h, k, l = reflection  # noqa: E741

        return self.volume / np.sqrt(
            self.s11 * h**2
            + self.s22 * k**2
            + self.s33 * l**2
            + 2 * (self.s12 * h * k + self.s23 * k * l + self.s31 * l * h)
        )

    @classmethod
    def from_name(cls, name: str) -> Material:
        """Initialize a material from its name."""
        if name in MATERIALS:
            return cls(*MATERIALS[name], name=name)
        raise ValueError(
            f"Unit cell parameters are not available for {name}. The available "
            f"materials are: {', '.join(MATERIALS.keys())}."
        )

    @classmethod
    def from_cif(cls, filename: str) -> Material:
        """Initialize a material from a CIF file."""
        try:
            from pymatgen.io.cif import CifParser
        except ImportError as e:
            raise ImportError(
                "Pymatgen is required to read CIF files. Follow the instructions at "
                "https://pymatgen.org/installation.html to install it."
            ) from e
        cif = CifParser(filename)
        structure = cif.get_structures()[0]
        return cls(
            *structure.lattice.abc, *structure.lattice.angles, name=structure.formula
        )



"""Bragg's law calculations."""

from __future__ import annotations

from typing import Iterable

import numpy as np
import numpy.typing as npt
from scipy.constants import Planck, elementary_charge, speed_of_light


def calculate_wavelength(
    angle: float | Iterable[float], d_spacing: float
) -> float | npt.NDArray[np.float64]:
    """
    Calculate the wavelength [Å] for the scattering angle [degrees] and d-spacing [Å].

    λ [Å] = 2 * d [Å] * sin(θ [rad])
    """
    angle = np.asarray(angle)
    return 2 * d_spacing * np.sin(np.radians(angle))


def calculate_energy(angle: float | Iterable[float], d_spacing: float):
    """
    Calculate the energy [keV] for the scattering angle [degrees] and d-spacing [Å].

    E [keV] = h [J s] * c [m s^-1] / e [C] / λ [Å] =  [eV m / Å] = 1e7 [keV]
    """
    wavelength = calculate_wavelength(angle, d_spacing)
    return 1e7 * Planck * speed_of_light / elementary_charge / wavelength


def calculate_scattering_angle(
    energy: float | Iterable[float], d_spacing: float
) -> float | npt.NDArray[np.float64]:
    """Calculate the angle [degrees] for a given energy [keV] and d-spacing [Å]."""
    energy = np.asarray(energy)
    wavelength = 1e7 * Planck * speed_of_light / elementary_charge / energy
    return np.degrees(np.arcsin(wavelength / (2 * d_spacing)))



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Iterator

if TYPE_CHECKING:
    from pathlib import Path

import h5py
import numpy as np
import pytest

from daxs.measurements import (
    ConcentrationCorrectionError,
    Measurement,
    Measurement1D,
    Rixs,
    SimpleConcentrationCorrector,
)
from daxs.scans import Scan, Scans
from daxs.sources import Hdf5Source
from daxs.utils import resources


def test_simple_concentration_correction_mock():
    # The same number of scans and points in the concentration correction scan.
    conc_corr_scan = Scan(None, np.array([0.5, 1.0]))
    corrector = SimpleConcentrationCorrector(conc_corr_scan)
    scans = [
        Scan(None, np.array([1.0, 2.0, 3.0])),
        Scan(None, np.array([4.0, 5.0, 6.0])),
    ]
    scans = Scans(scans)
    corrector.apply(scans)
    assert np.allclose(scans[0].signal, np.array([2, 4, 6]))
    assert np.allclose(scans[1].signal, np.array([4, 5, 6]))

    scans[0].monitor = np.array([10, 20, 30])
    conc_corr_scan.monitor = np.array([100, 200])
    corrector.apply(scans)
    assert np.allclose(scans[0].monitor, np.array([0.1, 0.2, 0.3]))

    # Different number of scans and points in the concentration correction scan.
    # The concentration correction scan has the same number of points as the scans.
    scans.reset()
    conc_corr_scan = Scan(None, np.array([0.5, 1.0, 2.0]))
    corrector = SimpleConcentrationCorrector(conc_corr_scan)
    corrector.apply(scans)
    assert np.allclose(scans[0].signal, np.array([2, 2, 1.5]))
    assert np.allclose(scans[1].signal, np.array([8, 5, 3]))

    # The number of scans and concentration correction scans is the same.
    scans.reset()
    conc_corr_scans = [
        Scan(None, np.array([0.5, 1.0, 2.0])),
        Scan(None, np.array([1.0, 2.0, 3.0])),
    ]
    conc_corr_scans = Scans(conc_corr_scans)
    corrector = SimpleConcentrationCorrector(conc_corr_scans)
    corrector.apply(scans)
    assert np.allclose(scans[0].signal, np.array([2, 2, 1.5]))
    assert np.allclose(scans[1].signal, np.array([4, 2.5, 2]))


@pytest.fixture
def mock_hdf5_path(tmp_path_factory: pytest.TempPathFactory) -> Iterator[Path]:
    path = tmp_path_factory.mktemp("files") / "tmp.h5"
    with h5py.File(path, "w") as fh:
        for i in range(1, 4):
            fh.create_dataset(f"{i}.1/measurement/x", data=[i, i + 1, i + 2])
            fh.create_dataset(f"{i}.1/measurement/signal", data=[i, i * 2, i * 3])
    yield path
    os.remove(path)


def test_measurement_init(mock_hdf5_path: str):
    data_mappings = {"x": ".1/measurement/x", "signal": ".1/measurement/signal"}
    source = Hdf5Source(mock_hdf5_path, None, None, data_mappings)

    with pytest.raises(ValueError):
        measurement = Measurement(source)
        _ = measurement.scans

    source = Hdf5Source(mock_hdf5_path, 1, None, data_mappings)
    measurement = Measurement(source)
    assert np.allclose(measurement.scans[0].signal, np.array([1, 2, 3]))

    source = Hdf5Source(mock_hdf5_path, [1, 2], None, data_mappings)
    measurement = Measurement(source)
    assert np.allclose(measurement.scans[1].x, np.array([2, 3, 4]))
    assert np.allclose(measurement.scans[1].signal, np.array([2, 4, 6]))


def test_measurement_remove_scans(mock_hdf5_path: str):
    data_mappings = {"x": ".1/measurement/x", "signal": ".1/measurement/signal"}
    source = Hdf5Source(mock_hdf5_path, [1, 2, 3], None, data_mappings)
    measurement = Measurement(source)

    indices = [1, 2, 3]
    measurement.remove_scans(indices)
    assert len(measurement.scans) == 0


@pytest.fixture()
def hdf5_path() -> str:
    return resources.getfile("Pd_foil_La_XANES.h5")


@pytest.fixture()
def data_mappings():
    return {
        "x": ".1/measurement/hdh_angle",
        "signal": [".1/measurement/g09", ".1/measurement/g14"],
    }


def test_measurement1d_properties(hdf5_path: str, data_mappings: dict[str, Any]):
    source = Hdf5Source(hdf5_path, [3, 4, 7], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    assert measurement.x[-1] == pytest.approx(38.72936736)
    assert measurement.signal[-1] == pytest.approx(21.1642857)

    data_mappings.update({"monitor": ".1/measurement/I0t"})
    source = Hdf5Source(hdf5_path, [3, 4, 7], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    assert measurement.monitor[-1] == pytest.approx(167717.885714)

    measurement.process(normalization="area")
    assert measurement.signal[-1] == pytest.approx(0.0012821686)

    measurement.x = np.append(measurement.x, 39.0)
    assert np.isnan(measurement.signal[-1])


def test_measurement1d_aggregate(hdf5_path: str, data_mappings: dict[str, Any]):
    source = Hdf5Source(hdf5_path, [7], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    assert np.all(measurement.x == measurement.scans[0].x)

    measurement.aggregate()
    assert np.all(measurement.signal == measurement.scans[0].signal)

    assert np.all(measurement.signal == measurement.scans[0].signal)

    data_mappings.update({"monitor": ".1/measurement/I0t"})
    source = Hdf5Source(hdf5_path, [7], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    measurement.aggregate()
    assert measurement.signal[-1] == pytest.approx(0.0001546175)

    source = Hdf5Source(hdf5_path, [3, 7], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    measurement.aggregate(mode="fraction of sums")
    assert measurement.signal[-1] == pytest.approx(0.0001306488)

    measurement.reset()
    measurement.aggregate(mode="sum of fractions")
    assert measurement.signal[-1] == pytest.approx(0.0002612987)

    with pytest.raises(ValueError):
        measurement.aggregate(mode="wrong")


def test_measurement1d_normalize(hdf5_path: str, data_mappings: dict[str, Any]):
    data_mappings.update({"monitor": ".1/measurement/I0t"})
    source = Hdf5Source(hdf5_path, [7], data_mappings=data_mappings)
    measurement = Measurement1D(source)

    with pytest.raises(ValueError):
        measurement.normalize(mode="wrong")

    measurement.normalize(mode="area")
    assert measurement.signal[-1] == pytest.approx(0.0015740913)

    measurement.normalize(mode="maximum")
    assert measurement.signal[-1] == pytest.approx(0.0004231201)

    measurement.reset()
    measurement.normalize(mode="maximum")
    assert measurement.signal[-1] == pytest.approx(0.0004231201)


def test_measurement1d_outliers(hdf5_path: str, data_mappings: dict[str, Any]):
    data_mappings.update({"monitor": ".1/measurement/I0t"})
    source = Hdf5Source(hdf5_path, [3, 5], data_mappings=data_mappings)
    measurement = Measurement1D(source)

    measurement.find_outliers(method="hampel", threshold=2.0)
    outliers = measurement.scans[0].outliers
    assert np.all(np.where(outliers)[1][:2] == [28, 60])

    measurement.reset()
    measurement.remove_outliers(method="hampel", threshold=3.5)
    outliers = measurement.scans[0].outliers
    assert np.all(np.where(outliers)[1][:2] == [123, 284])


def test_measurement1d_dead_time_correction(
    hdf5_path: str, data_mappings: dict[str, Any]
):
    data_mappings.update({"detection_time": ".1/measurement/sec"})
    source = Hdf5Source(hdf5_path, [3, 5], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    with pytest.raises(TypeError):
        measurement.dead_time_correction()  # type: ignore
    measurement.dead_time_correction(tau=[0.1, 0.1])


def test_measurement1d_save(hdf5_path: str, data_mappings: dict[str, Any]):
    source = Hdf5Source(hdf5_path, [3, 5], data_mappings=data_mappings)
    measurement = Measurement1D(source)
    with pytest.raises(TypeError):
        measurement.save()  # type: ignore
    filename = "test.dat"
    measurement.save(filename)
    assert os.path.isfile(filename)
    os.remove(filename)


@pytest.mark.parametrize(
    "scan_ids, conc_corr_ids, expected",
    (
        ([4], [5], [0.0750053, 0.06380908]),
        ([4, 10], [5, 11], [0.07468005, 0.06404491]),
        ([4], [6], ConcentrationCorrectionError),
        ([4, 10], [5, 11, 12], ConcentrationCorrectionError),
    ),
)
def test_simple_concentration_correction_equal_number_of_scans(
    scan_ids: list[int], conc_corr_ids: list[int], expected: Any
):
    hdf5_path: str = resources.getfile("A1_Kb_XES.h5")
    data_mappings = {
        "x": ".1/measurement/xes_en",
        "signal": ".1/measurement/det_dtc_apd",
        "monitor": ".1/measurement/I02",
    }

    source = Hdf5Source(hdf5_path, scan_ids, data_mappings=data_mappings)
    measurement = Measurement1D(source)
    if isinstance(expected, list):
        measurement.concentration_correction(conc_corr_ids)
        assert measurement.signal[-2:] == pytest.approx(expected)
    else:
        with pytest.raises(expected):
            measurement.concentration_correction(conc_corr_ids)


def test_simple_concentration_correction_equal_number_of_scans_and_points():
    hdf5_path = resources.getfile("Fe2O3_Ka1Ka2_RIXS.h5")
    data_mappings = {
        "x": ".1/measurement/zaptime",
        "y": ".1/instrument/positioners/xes_en",
        "signal": [".1/measurement/det_dtc_apd"],
        "monitor": ".1/measurement/I02",
    }
    source = Hdf5Source(hdf5_path, list(range(4, 225)), data_mappings=data_mappings)
    measurement = Rixs(source)
    measurement.concentration_correction([225])
    assert measurement.scans[0].signal[-1] == pytest.approx(0.01887687)
    assert measurement.scans[-1].monitor[0] == pytest.approx(0.24412638)
    assert measurement.signal.mean() == pytest.approx(0.02197076489)



from __future__ import annotations

import logging
import os
import sys
from collections.abc import MutableMapping
from typing import Any, Iterator

from ruamel.yaml import YAML

logger = logging.getLogger(__name__)


DEFAULT_CONFIG_FILENAME: str = "config.yml"
DEFAULT_SETTINGS: dict[str, Any] = {
    "use_blissdata_api": False,
    "dynamic_hdf5_retry_timeout": 0.1,
}


class Config(MutableMapping):
    """
    A singleton class to provide access to the configuration settings throughout
    the package. Implements dictionary-like interface for direct access to settings.
    """

    _instance: Config | None = None
    _settings: dict[str, Any] | None = None

    def __new__(cls) -> Config:
        """
        Create or return the singleton instance of the Config class.
        Ensures configuration is loaded when first instantiated.
        If no configuration file exists, creates one with default values.
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls._instance.read()

            # Check if any config file was loaded
            config_path = cls._instance._get_global_config_dir()
            config_file = os.path.join(config_path, DEFAULT_CONFIG_FILENAME)

            if not os.path.exists(config_file):
                # Create default configuration file if none exists
                cls._instance.write(default=True)

        return cls._instance

    def __getitem__(self, key: str) -> Any:
        """Get configuration value by key."""
        if self._settings is None:
            raise KeyError(key)
        return self._settings[key]

    def __setitem__(self, key: str, value: Any) -> None:
        """Set configuration value by key."""
        if self._settings is not None:
            self._settings[key] = value

    def __delitem__(self, key: str) -> None:
        """Delete configuration value by key."""
        if self._settings is not None:
            del self._settings[key]

    def __iter__(self) -> Iterator[str]:
        """Iterate through configuration keys."""
        if self._settings is not None:
            return iter(self._settings)
        return iter([])

    def __len__(self) -> int:
        """Get number of configuration items."""
        return len(self._settings) if self._settings is not None else 0

    def _get_global_config_dir(self) -> str:
        """
        Return the platform-specific directory path for storing the configuration file.

        Creates the directory if it doesn't exist and falls back to the current
        working directory if creation fails.

        Returns
        -------
            str: Path to the configuration directory
        """
        package = "daxs"
        if sys.platform.startswith("win"):
            base_dir = os.environ.get("APPDATA", "")
            if not base_dir:
                base_dir = os.path.join(
                    os.environ.get("USERPROFILE", ""), "AppData", "Roaming"
                )
            config_dir = os.path.join(base_dir, package)
        elif sys.platform.startswith("darwin"):
            base_dir = os.path.expanduser("~")
            config_dir = os.path.join(
                base_dir, "Library", "Application Support", package
            )
        else:
            xdg_config_home = os.environ.get("XDG_CONFIG_HOME", "")
            if not xdg_config_home:
                xdg_config_home = os.path.join(os.path.expanduser("~"), ".config")
            config_dir = os.path.join(xdg_config_home, package)

        # Create the config directory if it doesn't exist.
        if not os.path.exists(config_dir):
            try:
                os.makedirs(config_dir)
            except OSError as e:
                logger.info(f"Could not create config directory: {e}")
                # Fall back to current directory if the directory could not be created.
                return os.path.join(os.getcwd())

        return config_dir

    def read(self) -> None:
        """
        Read the configuration settings from YAML files.

        Checks multiple locations in order of priority:
        1. Path specified in DAXS_CONFIG environment variable (highest)
        2. config.yml in current working directory
        3. config.yml in platform-specific global configuration directory (lowest)

        Settings from higher priority files override those from lower priority files.
        Falls back to the internal default configuration if no files are found.
        """
        # Start with default configuration.
        self._settings = DEFAULT_SETTINGS.copy()

        locations: list[str | None] = [
            os.path.join(self._get_global_config_dir(), DEFAULT_CONFIG_FILENAME),
            os.path.join(os.getcwd(), DEFAULT_CONFIG_FILENAME),
            os.environ.get("DAXS_CONFIG"),
        ]

        # Filter out invalid locations.
        locations = [location for location in locations if location is not None]

        config_found = False
        for location in locations:
            try:
                yaml = YAML(typ="safe")
                with open(location, "r") as fh:
                    config = yaml.load(fh)
                    if config is not None:
                        self._settings.update(config)
                    logger.info(f"Loaded configuration from {location}")
                    config_found = True
                    return
            except (FileNotFoundError, PermissionError):
                continue

        if not config_found:
            logger.info("No configuration found, using internal defaults.")

    def write(self, default: bool = False) -> bool:
        """
        Write configuration to the global configuration file.

        Parameters
        ----------
        default : bool, optional
            If True, write the default configuration instead of the current one,
            but only if the file doesn't exist yet. Default is False.

        Returns
        -------
        bool
            True if the configuration was written successfully, False otherwise.
        """
        config_path = self._get_global_config_dir()
        config_file = os.path.join(config_path, DEFAULT_CONFIG_FILENAME)

        # When default is True, only write if the file doesn't exist.
        if default and os.path.exists(config_file):
            logger.info(f"Configuration file already exists at {config_file}.")
            return False

        # Determine which configuration to write.
        config_to_write = DEFAULT_SETTINGS if default else self._settings

        try:
            yaml = YAML()
            with open(config_file, "w") as file:
                yaml.dump(config_to_write, file)
            logger.info(f"Saved configuration to {config_file}")
            return True
        except (PermissionError, OSError) as e:
            logger.error(f"Failed to save configuration: {e}")
            return False

    def __str__(self) -> str:
        """
        Return a string representation of the current configuration.

        Returns
        -------
        str
            A formatted string showing all configuration keys and values.
        """
        if self._settings is None:
            return "No configuration was read."

        items = [f"{key}: {value}" for key, value in self._settings.items()]
        return "\n".join(items)



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Iterator

if TYPE_CHECKING:
    from pathlib import Path

import h5py
import numpy as np
import pytest

from daxs.sources import BlissPath, Hdf5Source, Selection


# TODO: Use collection instead of sample.
def test_bliss_path():
    bliss_path = BlissPath(
        root="/data/visitor",
        proposal="blc1234",
        beamline="id00",
        session="20240101",
        sample="sample_1",
        dataset="xanes_0001",
    )
    assert bliss_path.session == "20240101"
    assert bliss_path.filename == "sample_1_xanes_0001.h5"
    assert bliss_path.path == (
        "/data/visitor/blc1234/id00/20240101/RAW_DATA"
        "/sample_1/sample_1_xanes_0001/sample_1_xanes_0001.h5"
    )
    bliss_path.sample = "sample_2"
    assert bliss_path.path == (
        "/data/visitor/blc1234/id00/20240101/RAW_DATA"
        "/sample_2/sample_2_xanes_0001/sample_2_xanes_0001.h5"
    )

    path = (
        "/data/visitor/blc1234/id00/20240101/RAW_DATA"
        "/sample_1/sample_1_xanes_0001/sample_1_xanes_0001.h5"
    )
    bliss_path = BlissPath.from_path(path)
    assert bliss_path.path == path


def test_selection_invalid():
    with pytest.raises(ValueError):
        list(Selection("1-3"))
    with pytest.raises(ValueError):
        list(Selection([1, 2, "fscan"]))  # type: ignore


@pytest.mark.parametrize(
    "items, normalized_items",
    (
        (1, [1]),
        ([1, 2, 3], [1, 2, 3]),
        (np.array([1, 2, 3.0]), [1, 2, 3]),
    ),
)
def test_selection(items: int | str, normalized_items: list[int | str]):
    assert list(Selection(items)) == normalized_items


@pytest.mark.parametrize(
    "included_scan_ids, excluded_scan_ids, selected_scan_ids",
    (
        ([1, 2, 3, 4], None, [1, 2, 3, 4]),
        ([1, 2, 3], np.array([1, 2, 3]), []),
        (np.array([1, 2, 3]), None, [1, 2, 3]),
    ),
)
def test_hdf5_source_selected_scans(
    mock_hdf5_path: str,
    included_scan_ids: Any,
    excluded_scan_ids: Any,
    selected_scan_ids: list[int],
):
    source = Hdf5Source(
        mock_hdf5_path, included_scan_ids, excluded_scan_ids, data_mappings={}
    )
    assert source.selected_scan_ids == selected_scan_ids


@pytest.fixture
def mock_hdf5_path(tmp_path_factory: pytest.TempPathFactory) -> Iterator[Path]:
    path = tmp_path_factory.mktemp("files") / "mock_hdf5_filename.h5"
    with h5py.File(path, "w") as f:
        f.create_dataset("1.1/instrument/name", data="detector")
        f.create_dataset("2.1/measurement/x", data=[1, 2, 3])
        f.create_dataset("2.1/measurement/signal", data=[2, 2, 2])
        f.create_dataset("2.1/instrument/name", data="detector")
        f.create_dataset("3.1/measurement/x", data=[10, 20, 30])
        f.create_dataset("3.1/measurement/signal", data=[4.1, 5.1, 6.1])
        f.create_dataset("3.1/measurement/monitor", data=[7.0, 8.0, 9.0])
        f.create_dataset("3.1/measurement/sec", data=[1, 1, 1])
    yield path
    os.remove(path)


def test_hdf5_source_scans(mock_hdf5_path: str):
    data_mappings = {}
    source = Hdf5Source(mock_hdf5_path, 1, None, data_mappings)
    with pytest.raises(ValueError):
        assert source.scans

    data_mappings = {"x": ".1/measurement/x", "signal": ".1/measurement/signal"}
    source = Hdf5Source(mock_hdf5_path, 1, None, data_mappings)
    with pytest.raises(KeyError):
        assert source.scans

    data_mappings["name"] = ".1/instrument/name"
    source = Hdf5Source(mock_hdf5_path, 2, None, data_mappings)
    assert source.scans[0].data["name"] == b"detector"
    data_mappings.pop("name")

    data_mappings["monitor"] = ".1/measurement/monitor"
    source = Hdf5Source(mock_hdf5_path, 3, None, data_mappings)
    assert np.all(source.scans[0].monitor == np.array([7.0, 8.0, 9.0]))
    data_mappings.pop("monitor")

    data_mappings["detection_time"] = ".1/measurement/detection_time"
    source = Hdf5Source(mock_hdf5_path, 3, None, data_mappings)
    with pytest.raises(KeyError):
        assert source.scans

    data_mappings["detection_time"] = ".1/measurement/sec"
    source = Hdf5Source(mock_hdf5_path, 3, None, data_mappings)
    assert np.all(source.scans[0].data["detection_time"] == np.array([1.0, 1.0, 1.0]))
    data_mappings.pop("detection_time")

    source = Hdf5Source(mock_hdf5_path, [2, 3], None, data_mappings)
    scans = source.scans
    assert np.all(scans[0].x == np.array([1, 2, 3]))
    assert np.all(scans[1].x == np.array([10, 20, 30]))

    source = Hdf5Source(mock_hdf5_path, [9], None, data_mappings)
    with pytest.raises(KeyError):
        source.scans



"""The module provides classes to deal with different types of data sources."""

from __future__ import annotations

import logging
import os
from abc import ABC, abstractmethod
from typing import Any, Iterator

import numpy as np
import numpy.typing as npt

from daxs.scans import Scan, Scans

logger = logging.getLogger(__name__)


class BlissPath:
    def __init__(  # noqa
        self,
        root: str,
        proposal: str,
        beamline: str,
        session: str,
        sample: str,
        dataset: str,
        data_type: str = "RAW_DATA",
    ) -> None:
        self.root = root
        self.proposal = proposal
        self.beamline = beamline
        self.session = session
        self.sample = sample
        self.dataset = dataset
        self.data_type = data_type

    @property
    def collection(self) -> str:
        return f"{self.sample}_{self.dataset}"

    @property
    def filename(self) -> str:
        return f"{self.collection}.h5"

    @property
    def path(self) -> str:
        return os.path.join(
            self.root,
            self.proposal,
            self.beamline,
            self.session,
            self.data_type,
            self.sample,
            self.collection,
            self.filename,
        )

    @classmethod
    def from_path(cls, path: str) -> BlissPath:
        """Create a BlissPath object from a path."""
        tokens: list[str] = os.path.normpath(path).split(os.sep)
        if not tokens:
            raise ValueError("Invalid path.")
        tokens = tokens[::-1]
        _, collection, sample, data_type, session, beamline, proposal, *root = tokens
        # Determine the dataset name.
        dataset = collection.split(sample)[1][1:]
        # Create the root.
        root = os.path.join(os.sep, *root[::-1])
        return cls(root, proposal, beamline, session, sample, dataset, data_type)


class Selection:
    def __init__(
        self,
        items: int | str | list[int] | npt.NDArray[np.int64 | np.float64],
    ) -> None:
        """
        Class to handle selections of items.

        Parameters
        ----------
        items :
            The items to select.

        """
        if not isinstance(items, (int, str, list, np.ndarray)):  # type: ignore
            raise ValueError(
                "The items must be an integer, a string, a list of integers, or a "
                "Numpy array. \n"
                "For example: 1, [1, 2, 3], np.array([1, 2, 3]), "
                "or the output of certain functions like np.arange(1, 10, 2). \n"
                f"Provided items: {items}"
            )
        self._items = items

    @property
    def items(self):
        if isinstance(self._items, int):
            return [self._items]
        elif isinstance(self._items, str):
            raise ValueError(
                (
                    "Selecting scans using strings is not supported anymore. ",
                    "Please use a list of integers or a Numpy array",
                )
            )
        elif isinstance(self._items, list):
            for item in self._items:
                if not isinstance(item, int):  # type: ignore
                    raise ValueError("The items must be integers.")
            return self._items
        elif isinstance(self._items, np.ndarray):  # type: ignore
            return self._items.astype(np.int64).tolist()
        else:
            raise ValueError(
                "The items must be an integer, list of integers, or Numpy array."
            )

    def __iter__(self) -> Iterator[int]:
        return iter(self.items)


class Source(ABC):
    """Base class for sources of scans."""

    @property
    @abstractmethod
    def filename(self) -> str | None:
        """The filename of the source."""

    @property
    @abstractmethod
    def data_mappings(self) -> dict[str, Any]:
        """The mappings between scan attributes and paths in the source."""

    @property
    @abstractmethod
    def scans(self) -> Scans:
        """Return all source scans."""

    @data_mappings.setter
    def data_mappings(self, data_mappings: dict[str, Any]) -> None:
        """Set the mappings between scan attributes and paths in the source."""

    @abstractmethod
    def read_scans(
        self, indices: list[int] | npt.NDArray[np.int64] | None = None
    ) -> Scans:
        """Return all source scans."""


class Hdf5Source(Source):
    def __init__(
        self,
        filename: str,
        included_scan_ids: Any = None,
        excluded_scan_ids: Any = None,
        data_mappings: dict[str, Any] | None = None,
    ) -> None:
        """
        Class for a HDF5 source of scans

        Parameters
        ----------
        filename :
            Name of the HDF5 file.
        included_scans_ids :
            Selection of included scans.
        excluded_scans_ids :
            Selection of excluded scans.
        data_mappings :
            Mappings between scan attributes (x, signal, monitor, etc.) and paths in
            the HDF5 file.

        """
        included_scan_ids = [] if included_scan_ids is None else included_scan_ids
        excluded_scan_ids = [] if excluded_scan_ids is None else excluded_scan_ids

        self._filename = filename
        self.included_scan_ids = Selection(included_scan_ids)
        self.excluded_scan_ids = Selection(excluded_scan_ids)

        if not isinstance(data_mappings, dict):
            raise ValueError("The data_mappings must be a dict.")
        self._data_mappings = data_mappings

    @property
    def filename(self) -> str:
        return self._filename

    @property
    def data_mappings(self) -> dict[str, Any]:
        return self._data_mappings

    @data_mappings.setter
    def data_mappings(self, data_mappings: dict[str, Any]) -> None:
        self._data_mappings = data_mappings

    @property
    def selected_scan_ids(self) -> list[int]:
        """Get the selected scans considering the included and excluded scan ids."""
        selected_scan_ids = []

        for index in sorted(self.included_scan_ids):
            if index not in self.excluded_scan_ids:
                selected_scan_ids.append(index)
            else:
                logger.info("Scan %s/%d was excluded.", self.filename, index)

        logger.debug(
            "The scans %s have been selected from %s.", selected_scan_ids, self.filename
        )
        return selected_scan_ids

    @property
    def scans(self) -> Scans:
        """Return all source scans."""
        return self.read_scans()

    def read_scans(
        self, indices: list[int] | npt.NDArray[np.int64] | None = None
    ) -> Scans:
        """Read the scans from the source."""
        if indices is None:
            indices = self.selected_scan_ids
        return Scans([self.read_scan(index) for index in indices])

    def read_scan(self, index: int) -> Scan:
        """Return a scan object at the index."""
        if "x" not in self.data_mappings:
            raise ValueError("The data_mappings attribute must contain an entry for x.")
        if "signal" not in self.data_mappings:
            raise ValueError(
                "The data_mappings attribute must contain an entry for signal."
            )
        scan = Scan()
        scan.filename = self.filename
        scan.index = index

        data = {}
        for key, data_paths in self.data_mappings.items():
            data[key] = scan.read_data_at_paths(data_paths)
        scan.data = data

        return scan
